/*
 * Copyright (c) 2016, Yuichiro MORIGUCHI
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * * Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * * Neither the name of the Yuichiro MORIGUCHI nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Yuichiro MORIGUCHI BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
using System;
using System.IO;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using System.Runtime.Serialization;

namespace Morilib.Awk
{
	public class Graph<T, L>
	{
		protected Dictionary<T, List<Edge<T, L>>> vertexes = new Dictionary<T, List<Edge<T, L>>>();
		public void Add(T vertex, Edge<T, L> edge)
		{
			List<Edge<T, L>> edges;
			if(vertexes.ContainsKey(vertex))
			{
				edges = vertexes[vertex];
			}
			else
			{
				edges = new List<Edge<T, L>>();
				vertexes[vertex] = edges;
			}
			edges.Add(edge);
		}
	}

	public class Edge<T, L>
	{
		public L Label { get; private set; }
		public T From { get; private set; }
		public T To { get; private set; }
		public Edge(L l, T f, T t)
		{
			Label = l;
			From = f;
			To = t;
		}
	}

	public delegate bool DelegateChrRange(int ch);
	public interface ChrRange
	{
		bool Contains(int ch);
	}

	public class ChrInterval : ChrRange
	{
		private int ch1, ch2;
		public ChrInterval(int ch1, int ch2)
		{
			this.ch1 = ch1;
			this.ch2 = ch2;
		}
		public bool Contains(int ch)
		{
			return ch >= ch1 && ch <= ch2;
		}
	}

	public class ChrRangeWrapper : ChrRange
	{
		private DelegateChrRange range;
		public ChrRangeWrapper(DelegateChrRange r)
		{
			range = r;
		}
		public bool Contains(int ch)
		{
			return range(ch);
		}
	}
		
	public class ChrSet
	{
		private List<ChrRange> ranges;
		private bool complement;
		public ChrSet(List<ChrRange> r, bool c)
		{
			ranges = new List<ChrRange>(r);
			complement = c;
		}
		public ChrSet(int ch1, int ch2)
		{
			ranges = new List<ChrRange>();
			complement = false;
			ranges.Add(new ChrInterval(ch1, ch2));
		}
		public ChrSet(int c) : this(c, c) {}
		public bool Contains(int ch)
		{
			if(complement && ch >= 0)
			{
				foreach(ChrRange ci in ranges)
				{
					if(ci.Contains(ch))
					{
						return false;
					}
				}
				return true;
			}
			else
			{
				foreach(ChrRange ci in ranges)
				{
					if(ci.Contains(ch))
					{
						return true;
					}
				}
				return false;
			}
		}
	}

	public class AwkRegexUtil
	{
		public static void AddAll(HashSet<int> dest, HashSet<int> src)
		{
			foreach(int s in src)
			{
				dest.Add(s);
			}
		}
	}

	public class NFA : Graph<int, ChrSet>
	{
		public const int Caret = -1;
		public const int Dollar = -2;
		public int Start { get; set; }
		public int Accept { get; set; }
		public HashSet<int> TransitEpsilon(int s)
		{
			HashSet<int> res = new HashSet<int>();
			HashSet<int> visited = new HashSet<int>();
			List<int> sts = new List<int>();
			int sn;
			sts.Add(s);  res.Add(s);
			while(sts.Count > 0)
			{
				sn = sts[sts.Count - 1];
				sts.RemoveAt(sts.Count - 1);
				if(!visited.Contains(sn))
				{
					if(vertexes.ContainsKey(sn))
					{
						foreach(Edge<int, ChrSet> eg in vertexes[sn])
						{
							if(eg.Label == null)
							{
								res.Add(eg.To);
								sts.Add(eg.To);
							}
						}
					}
					visited.Add(sn);
				}
			}
			return res;
		}
		public HashSet<int> TransitEpsilon(HashSet<int> sts)
		{
			HashSet<int> res = new HashSet<int>();
			foreach(int s in sts)
			{
				AwkRegexUtil.AddAll(res, TransitEpsilon(s));
			}
			return res;
		}
		public HashSet<int> Transit(int s, int ch)
		{
			HashSet<int> res = new HashSet<int>();
			if(vertexes.ContainsKey(s))
			{
				foreach(Edge<int, ChrSet> eg in vertexes[s])
				{
					if(eg.Label != null && eg.Label.Contains(ch))
					{
						res.Add(eg.To);
					}
				}
			}
			return res;
		}
		public HashSet<int> Transit(HashSet<int> sts, int ch)
		{
			HashSet<int> res = new HashSet<int>();
			foreach(int s in sts)
			{
				AwkRegexUtil.AddAll(res, Transit(s, ch));
			}
			return res;
		}
	}

	public class AwkRegex
	{
		private NFA nfa;
		public string Pattern { get; private set; }
		public AwkRegex(string pat)
		{
			Pattern = pat;
			if(pat.Equals(""))
			{
				nfa = null;
			}
			else
			{
				nfa = AwkRegexParser.ParsePattern(pat);
			}
		}
		public int Match(string str, int pos)
		{
			HashSet<int> sts = new HashSet<int>();
			int matched = -1;
			if(nfa == null)
			{
				return pos;
			}
			sts.Add(nfa.Start);
			sts = nfa.TransitEpsilon(sts);
			if(pos == 0)
			{
				AwkRegexUtil.AddAll(sts, nfa.Transit(sts, NFA.Caret));
				sts = nfa.TransitEpsilon(sts);
			}
			for(int p = pos; p < str.Length; p++)
			{
				if(sts.Contains(nfa.Accept))
				{
					matched = p;
				}
				sts = nfa.Transit(sts, str[p]);
				sts = nfa.TransitEpsilon(sts);
				if(sts.Count == 0)  // dead state
				{
					return matched;
				}
			}
			AwkRegexUtil.AddAll(sts, nfa.Transit(sts, NFA.Dollar));
			sts = nfa.TransitEpsilon(sts);
			return sts.Contains(nfa.Accept) ? str.Length : matched;
		}
		public string Replace(string input, string replacement, int count, ref int len)
		{
			StringBuilder b = new StringBuilder();
			int np;
			for(int i = 0; i < input.Length;)
			{
				if((count <= 0 || len < count) && (np = Match(input, i)) >= 0)
				{
					b.Append(replacement);
					i = np;
					len++;
				}
				else
				{
					b.Append(input[i]);
					i++;
				}
			}
			return b.ToString();
		}
		public AwkMatch Match(string input)
		{
			return new AwkMatch(this, input, 0);
		}
		public bool IsMatch(string input)
		{
			return Match(input).Success;
		}
		public string[] Split(string input)
		{
			List<string> res = new List<string>();
			int ind = 0;
			if(nfa != null)
			{
				for(AwkMatch mt = Match(input); mt.Success; mt = mt.NextMatch())
				{
					if(mt.Length > 0)
					{
						res.Add(input.Substring(ind, mt.Index - ind));
						ind = mt.Index + mt.Length;
					}
				}
				res.Add(input.Substring(ind, input.Length - ind));
			}
			else
			{
				for(int i = 0; i < input.Length; i++)
				{
					res.Add(input.Substring(i, 1));
				}
			}
			return res.ToArray();
		}
	}

	public class AwkMatch
	{
		public AwkRegex Regex { get; private set; }
		public string Input { get; private set; }
		public int Index { get; private set; }
		public int Length { get; private set; }
		public bool Success { get; private set; }
		public string Value { get; private set; }
		public AwkMatch(AwkRegex r, string s, int pos)
		{
			Regex = r;
			Input = s;
			GetNext(pos);
		}
		private void GetNext(int pos)
		{
			int np;
			for(int i = pos; i <= Input.Length; i++)
			{
				if((np = Regex.Match(Input, i)) >= 0)
				{
					Success = true;
					Index = i;
					Length = np - i;
					Value = Input.Substring(i, Length);
					return;
				}
			}
			Success = false;
			Index = -1;
			Length = -1;
			Value = null;
		}
		public AwkMatch NextMatch()
		{
			if(Success)
			{
				return new AwkMatch(Regex, Input, Index + (Length > 0 ? Length : 1));
			}
			else
			{
				return null;
			}
		}
	}

}