/*
 * Copyright (c) 2016, Yuichiro MORIGUCHI
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * * Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * * Neither the name of the Yuichiro MORIGUCHI nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Yuichiro MORIGUCHI BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
using System;
using System.IO;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using System.Runtime.Serialization;

namespace Morilib.Awk
{
	public class AwkRuntimeException : Exception
	{
		public AwkRuntimeException() : base()
		{
		}
		public AwkRuntimeException(string message) : base(message)
		{
		}
		public AwkRuntimeException(string message, Exception inner) : base(message, inner)
		{
		}

		protected AwkRuntimeException(SerializationInfo info, StreamingContext context)
        {
        }
	}

	public class Env
	{
		public AwkStream Stream { get; set; }
		private int maxfield;
		private Dictionary<int, string> fields = new Dictionary<int, string>();
		private Dictionary<string, object> variables = new Dictionary<string, object>();
		private Dictionary<string, Function> functions = new Dictionary<string, Function>();
		private Dictionary<string, StreamWriter> writers = new Dictionary<string, StreamWriter>();
		private Dictionary<string, AwkStream> readers = new Dictionary<string, AwkStream>();
		public Env()
		{
			variables["FS"] = new Str(" ");
			variables["RS"] = new Str("\n");
			variables["OFS"] = new Str(" ");
			variables["ORS"] = new Str("\n");
			variables["SUBSEP"] = new Str("\u0034");
			variables["RSTART"] = new Num(0);
			variables["RLENGTH"] = new Num(0);
			variables["ENCODE"] = new Str(Encoding.Default.WebName);
			variables["CONVFMT"] = new Str("%.6g");
			variables["OFMT"] = new Str("%.6g");
			variables["ENVIRON"] = GetEnvironment();
			variables["FNR"] = new Num(0);
			variables["NF"] = new Num(0);
			variables["NR"] = new Num(0);
			functions["int"] = Functions.Fint;
			functions["sqrt"] = Functions.Fsqrt;
			functions["exp"] = Functions.Fexp;
			functions["log"] = Functions.Flog;
			functions["sin"] = Functions.Fsin;
			functions["cos"] = Functions.Fcos;
			functions["atan2"] = Functions.Fatan2;
			functions["rand"] = Functions.Frand;
			functions["srand"] = Functions.Fsrand;
			functions["index"] = Functions.Findex;
			functions["length"] = Functions.Flength;
			functions["match"] = Functions.Fmatch;
			functions["substr"] = Functions.Fsubstr;
			functions["tolower"] = Functions.Ftolower;
			functions["toupper"] = Functions.Ftoupper;
			functions["sprintf"] = Functions.Fsprintf;
			functions["close"] = Functions.Fclose;
		}
		public Env(List<string> f) : this()
		{
			SetField(f);
		}
		public Env(AwkStream s) : this()
		{
			Stream = s;
		}
		private Dictionary<string, Value> GetEnvironment()
		{
			Dictionary<string, Value> res = new Dictionary<string, Value>();
			foreach(DictionaryEntry de in Environment.GetEnvironmentVariables())
			{
				res[de.Key.ToString()] = new Str(de.Value.ToString());
			}
			return res;
		}
		public void SetField(List<string> f)
		{
			fields = new Dictionary<int, string>();
			for(int i = 0; i < f.Count; i++)
			{
				fields.Add(i, f[i]);
			}
			maxfield = f.Count - 1;
			variables["NF"] = new Num(maxfield);
		}
		public string GetField(int num)
		{
			if(num < 0)
			{
				throw new Exception();
			}
			else if(!fields.ContainsKey(num))
			{
				return "";
			}
			else
			{
				return fields[num];
			}
		}
		private bool IsFSMatch(int ch, string fs)
		{
			if(fs[0] == ' ')
			{
				return " \t\n".IndexOf((char)ch) >= 0;
			}
			else
			{
				return ch == fs[0];
			}
		}
		public void SetField(int num, Value val)
		{
			StringBuilder bl;
			List<string> f;
			string[] sp;
			string dm = "";
			string fs;
			int p;
			bool mt = false;
			if(num < 0)
			{
				throw new AwkRuntimeException("field out of range");
			}
			else if(num == 0)
			{
				f = new List<string>();
				f.Add(val.ToStr(this));
				fs = GetFS();
				if(fs.Length == 1)
				{
					bl = new StringBuilder();
					p = 0;
					for(; p < f[0].Length; p++)
					{
						if(mt)
						{
							if(IsFSMatch(f[0][p], fs))
							{
								f.Add(bl.ToString());
								bl = new StringBuilder();
								mt = false;
							}
							else
							{
								bl.Append(f[0][p]);
							}
						}
						else
						{
							if(!IsFSMatch(f[0][p], fs))
							{
								bl.Append(f[0][p]);
								mt = true;
							}
						}
					}
					dm = bl.ToString();
					if(dm.Length > 0)
					{
						f.Add(dm);
					}
					SetField(f);
				}
				else
				{
					sp = new AwkRegex(fs).Split(f[0]);
					foreach(string s in sp)
					{
						f.Add(s);
					}
					SetField(f);
				}
			}
			else
			{
				fields[num] = val.ToStr(this);
				if(num > maxfield)
				{
					maxfield = num;
					variables["NF"] = new Num(num);
				}
				bl = new StringBuilder();
				for(int i = 1; i <= maxfield; i++)
				{
					bl.Append(dm);
					bl.Append(GetField(i));
					if(dm.Equals(""))
					{
						dm = ((Value)variables["OFS"]).ToStr(this);
					}
				}
				fields[0] = bl.ToString();
			}
		}
		public void SetVariable(string name, Value val)
		{
			if(!variables.ContainsKey(name) || variables[name] is Value)
			{
				variables[name] = val;
			}
			else
			{
				throw new AwkRuntimeException(name + " is an array name");
			}
		}
		public Value GetVariable(string name)
		{
			if(!variables.ContainsKey(name))
			{
				return new Str("");
			}
			else if(variables[name] is Value)
			{
				return (Value)variables[name];
			}
			else
			{
				throw new AwkRuntimeException(name + " is an array name");
			}
		}
		public string GetFS()
		{
			return ((Value)variables["FS"]).ToStr(this);
		}
		public int GetRS()
		{
			string rs;
			rs = ((Value)variables["RS"]).ToStr(this);
			return rs.Equals("") ? -1 : rs[0];
		}
		public string GetConvfmt()
		{
			return ((Value)variables["CONVFMT"]).ToStr(this);
		}
		public string GetOfmt()
		{
			return ((Value)variables["OFMT"]).ToStr(this);
		}
		public string GetOFS()
		{
			return ((Value)variables["OFS"]).ToStr(this);
		}
		public string GetORS()
		{
			return ((Value)variables["ORS"]).ToStr(this);
		}
		public void IncNR()
		{
			variables["NR"] = new Num(((Value)variables["NR"]).ToNum(this) + 1);
			variables["FNR"] = new Num(((Value)variables["FNR"]).ToNum(this) + 1);
		}
		public void ResetFNR()
		{
			variables["FNR"] = new Num(0);
		}
		public void SetArgv(List<string> argv)
		{
			Dictionary<string, Value> ar = new Dictionary<string, Value>();
			variables["ARGC"] = new Num(argv.Count);
			for(int i = 0; i < argv.Count; i++)
			{
				ar[i.ToString()] = new Str(argv[i]);
			}
			variables["ARGV"] = ar;
		}
		private string ToArg(List<Value> vs)
		{
			StringBuilder arg = new StringBuilder();
			string dl = "";
			foreach(Value v in vs)
			{
				arg.Append(dl);
				arg.Append(v.ToStr(this));
				dl = ((Value)variables["SUBSEP"]).ToStr(this);
			}
			return arg.ToString();
		}
		public void SetArray(string name, List<Value> vs, Value v1)
		{
			Dictionary<string, Value> nar;
			string arg;
			if(variables.ContainsKey(name) && variables[name] is Value)
			{
				throw new AwkRuntimeException(name + " is an array name");
			}
			arg = ToArg(vs);
			if(variables.ContainsKey(name))
			{
				((Dictionary<string, Value>)variables[name])[arg.ToString()] = v1;
			}
			else
			{
				nar = new Dictionary<string, Value>();
				nar[arg.ToString()] = v1;
				variables[name] = nar;
			}
		}
		public Value GetArray(string name, List<Value> vs)
		{
			Dictionary<string, Value> nar;
			string arg;
			if(variables.ContainsKey(name) && variables[name] is Value)
			{
				throw new AwkRuntimeException(name + " is an array name");
			}
			arg = ToArg(vs);
			if(!variables.ContainsKey(name))
			{
				return new Str("");
			}
			else
			{
				nar = (Dictionary<string, Value>)variables[name];
				if(nar.ContainsKey(arg))
				{
					return nar[arg];
				}
				else
				{
					return new Str("");
				}
			}
		}
		public bool InArray(string name, List<Value> vs)
		{
			Dictionary<string, Value> nar;
			string arg;
			if(!variables.ContainsKey(name))
			{
				return false;
			}
			else if(variables[name] is Value)
			{
				throw new AwkRuntimeException(name + " is not an array name");
			}
			else
			{
				arg = ToArg(vs);
				nar = (Dictionary<string, Value>)variables[name];
				return nar.ContainsKey(arg);
			}
		}
		public List<Value> GetArrays(string name)
		{
			Dictionary<string, Value> nar;
			List<Value> res = new List<Value>();
			if(!variables.ContainsKey(name))
			{
			}
			else if(variables[name] is Value)
			{
				throw new AwkRuntimeException(name + " is not an array name");
			}
			else
			{
				nar = (Dictionary<string, Value>)variables[name];
				foreach(string v in nar.Keys)
				{
					res.Add(new Str(v));
				}
			}
			return res;
		}
		public void ClearArray(string name)
		{
			if(!variables.ContainsKey(name))
			{
				// do nothing
			}
			else if(variables[name] is Value)
			{
				throw new AwkRuntimeException(name + " is not an array name");
			}
			else
			{
				variables.Remove(name);
			}
		}
		public Value CallFunction(string name, List<Value> vs)
		{
			Function f;
			if(!functions.ContainsKey(name))
			{
				throw new AwkRuntimeException("function " + name + " is not defined");
			}
			f = functions[name];
			return f(this, vs);
		}
		public StreamWriter OpenWriter(string name, bool append)
		{
			Encoding enc;
			if(!writers.ContainsKey(name))
			{
				enc = Encoding.GetEncoding(((Value)variables["ENCODE"]).ToStr(this));
				writers[name] = new StreamWriter(name, append, enc);
			}
			return writers[name];
		}
		public AwkStream OpenAwkStream(string name)
		{
			StreamReader rd;
			Encoding enc;
			enc = Encoding.GetEncoding(((Value)variables["ENCODE"]).ToStr(this));
			rd = new StreamReader(name, enc);
			return new AwkStream(rd);
		}
		public AwkStream OpenReader(string name)
		{
			if(!readers.ContainsKey(name))
			{
				readers[name] = OpenAwkStream(name);
			}
			return readers[name];
		}
		public bool Close(string name)
		{
			if(writers.ContainsKey(name))
			{
				writers[name].Close();
				writers.Remove(name);
				return true;
			}
			else if(readers.ContainsKey(name))
			{
				readers[name].Close();
				readers.Remove(name);
				return true;
			}
			else
			{
				return false;
			}
		}
		public void CloseAll()
		{
			foreach(StreamWriter w in writers.Values)
			{
				w.Close();
			}
		}
	}

	public delegate bool UnaryBool(Env e, Value v1);
	public delegate int UnaryInt(Env e, Value v1);
	public delegate Value Unary(Env e, Value v1);
	public delegate bool BinaryBool(Env e, Value v1, Value v2);
	public delegate Value Binary(Env e, Value v1, Value v2);

	public delegate Value Function(Env e, List<Value> args);

	public class Functions
	{
		private static Random rand = new Random(765);
		private static void CheckArity(string fname, List<Value> args, int arity)
		{
			if(args.Count != arity)
			{
				throw new AwkRuntimeException(fname + ": invalid arity");
			}
		}
		private static void CheckArity(string fname, List<Value> args, int arityb, int aritye)
		{
			if(args.Count < arityb || args.Count > aritye)
			{
				throw new AwkRuntimeException(fname + ": invalid arity");
			}
		}
		private static void CheckArityLess(string fname, List<Value> args, int arity)
		{
			if(args.Count > arity)
			{
				throw new AwkRuntimeException(fname + ": invalid arity");
			}
		}
		private static void CheckArityMore(string fname, List<Value> args, int arity)
		{
			if(args.Count < arity)
			{
				throw new AwkRuntimeException(fname + ": invalid arity");
			}
		}
		public static Value Fint(Env e, List<Value> args)
		{
			double num;
			CheckArity("int", args, 1);
			num = args[0].ToNum(e);
			num = num > 0 ? Math.Floor(num) : Math.Ceiling(num);
			return new Num(num);
		}
		public static Value Fsqrt(Env e, List<Value> args)
		{
			double num;
			CheckArity("sqrt", args, 1);
			num = args[0].ToNum(e);
			return new Num(Math.Sqrt(num));
		}
		public static Value Fexp(Env e, List<Value> args)
		{
			double num;
			CheckArity("exp", args, 1);
			num = args[0].ToNum(e);
			return new Num(Math.Exp(num));
		}
		public static Value Flog(Env e, List<Value> args)
		{
			double num;
			CheckArity("log", args, 1);
			num = args[0].ToNum(e);
			return new Num(Math.Log(num));
		}
		public static Value Fsin(Env e, List<Value> args)
		{
			double num;
			CheckArity("sin", args, 1);
			num = args[0].ToNum(e);
			return new Num(Math.Sin(num));
		}
		public static Value Fcos(Env e, List<Value> args)
		{
			double num;
			CheckArity("cos", args, 1);
			num = args[0].ToNum(e);
			return new Num(Math.Cos(num));
		}
		public static Value Fatan2(Env e, List<Value> args)
		{
			double num1, num2;
			CheckArity("atan2", args, 2);
			num1 = args[0].ToNum(e);
			num2 = args[1].ToNum(e);
			return new Num(Math.Atan2(num1, num2));
		}
		public static Value Frand(Env e, List<Value> args)
		{
			CheckArity("rand", args, 0);
			return new Num(rand.NextDouble());
		}
		public static Value Fsrand(Env e, List<Value> args)
		{
			double num;
			CheckArityLess("srand", args, 1);
			if(args.Count > 0)
			{
				num = args[0].ToNum(e);
				rand = new Random((int)num);
			}
			else
			{
				rand = new Random();
			}
			return new Num(0);
		}
		public static Value Findex(Env e, List<Value> args)
		{
			string s1, s2;
			CheckArity("index", args, 2);
			s1 = args[0].ToStr(e);
			s2 = args[1].ToStr(e);
			return new Num(s1.IndexOf(s2) + 1);
		}
		public static Value Flength(Env e, List<Value> args)
		{
			string s1;
			CheckArityLess("length", args, 1);
			s1 = args.Count > 0 ? args[0].ToStr(e) : e.GetField(0);
			return new Num(s1.Length);
		}
		public static Value Fmatch(Env e, List<Value> args)
		{
			AwkRegex r1;
			string s1;
			AwkMatch m;
			CheckArity("match", args, 2);
			s1 = args[0].ToStr(e);
			r1 = args[1].ToRE(e);
			m = r1.Match(s1);
			if(m.Success)
			{
				e.SetVariable("RSTART", new Num(m.Index + 1));
				e.SetVariable("RLENGTH", new Num(m.Length));
				return new Num(m.Index);
			}
			else
			{
				e.SetVariable("RSTART", new Num(0));
				e.SetVariable("RLENGTH", new Num(-1));
				return new Num(0);
			}
		}
		public static Value Fsubstr(Env e, List<Value> args)
		{
			string s1;
			int id, ln;
			CheckArity("substr", args, 2, 3);
			s1 = args[0].ToStr(e);
			id = (int)args[1].ToNum(e) - 1;
			ln = args.Count > 2 ? (int)args[2].ToNum(e) : s1.Length - id;
			if(s1.Length == 0)
			{
				return new Str("");
			}
			else if(id < 0)
			{
				throw new AwkRuntimeException("substr: index must be positive");
			}
			else if(ln < 0)
			{
				throw new AwkRuntimeException("substr: length must be non negative");
			}
			else if(ln > s1.Length - id)
			{
				ln = s1.Length - id;
			}
			return new Str(s1.Substring(id, ln));
		}
		public static Value Ftolower(Env e, List<Value> args)
		{
			string s1;
			CheckArity("tolower", args, 1);
			s1 = args[0].ToStr(e);
			return new Str(s1.ToLower());
		}
		public static Value Ftoupper(Env e, List<Value> args)
		{
			string s1;
			CheckArity("toupper", args, 1);
			s1 = args[0].ToStr(e);
			return new Str(s1.ToUpper());
		}
		public static Value Fsprintf(Env e, List<Value> args)
		{
			string s1;
			Value[] va = new Value[args.Count - 1];
			CheckArityMore("sprintf", args, 1);
			s1 = args[0].ToStr(e);
			for(int i = 1; i < args.Count; i++)
			{
				va[i - 1] = args[i];
			}
			return new Str(AwkPrintf.Printf(e, s1, va));
		}
		public static Value Fclose(Env e, List<Value> args)
		{
			string s1;
			CheckArity("close", args, 1);
			s1 = args[0].ToStr(e);
			return new Num(e.Close(s1) ? 1 : 0);
		}
	}

	public abstract class Value
	{
		public abstract bool ToBool(Env e);
		public abstract bool ToBoolMatch(Env e);
		public abstract string ToStr(Env e);
		public abstract double ToNum(Env e);
		public abstract Value ToRval(Env e);
		public abstract AwkRegex ToRE(Env e);
		public abstract bool IsEqualTo(Value v);
		public abstract bool IsNum(Env e);
		public bool IsMatch(Env e, Value v)
		{
			return v.ToRE(e).IsMatch(this.ToStr(e));
		}
		public int CompareTo(Env e, Value v)
		{
			double v1, v2;
			string s1, s2;
			int cp;
			if(this.IsNum(e) && v.IsNum(e))
			{
				v1 = this.ToNum(e);
				v2 = v.ToNum(e);
				return v1 < v2 ? -1 : v1 > v2 ? 1 : 0;
			}
			else
			{
				s1 = this.ToStr(e);
				s2 = v.ToStr(e);
				cp = s1.CompareTo(s2);
				return cp < 0 ? -1 : cp > 0 ? 1 : 0;
			}
		}
		public bool EqualTo(Env e, Value v)
		{
			return this.ToStr(e).Equals(v.ToStr(e));
		}
		public Value Concat(Env e, Value v)
		{
			return new Str(this.ToStr(e) + v.ToStr(e));
		}
		public Value Add(Env e, Value v)
		{
			return new Num(this.ToNum(e) + v.ToNum(e));
		}
		public Value Sub(Env e, Value v)
		{
			return new Num(this.ToNum(e) - v.ToNum(e));
		}
		public Value Mul(Env e, Value v)
		{
			return new Num(this.ToNum(e) * v.ToNum(e));
		}
		public Value Div(Env e, Value v)
		{
			double v2 = v.ToNum(e);
			if(v2 == 0)
			{
				throw new AwkRuntimeException("division by zero");
			}
			return new Num(this.ToNum(e) / v2);
		}
		public Value Mod(Env e, Value v)
		{
			double v2 = v.ToNum(e), rs;
			if(v2 == 0)
			{
				throw new AwkRuntimeException("division by zero");
			}
			rs = Math.IEEERemainder(this.ToNum(e), v2);
			return new Num(rs < 0 ? rs + v2 : rs);
		}
		public Value Neg(Env e)
		{
			return new Num(-this.ToNum(e));
		}
		public Value Pos(Env e)
		{
			return new Num(this.ToNum(e));
		}
		public Value Pow(Env e, Value v)
		{
			return new Num(Math.Pow(this.ToNum(e), v.ToNum(e)));
		}
		public Value Field(Env e)
		{
			double v = this.ToNum(e);
			if(v < 0 || v > 10000)
			{
				throw new AwkRuntimeException("field out of range");
			}
			return new Str(e.GetField((int)v));
		}
		public Value Not(Env e)
		{
			return new Num(this.ToBool(e) ? 0 : 1);
		}

		public static bool IsMatch(Env e, Value w, Value v)
		{
			return w.IsMatch(e, v);
		}
		public static int CompareTo(Env e, Value w, Value v)
		{
			return w.CompareTo(e, v);
		}
		public static bool EqualTo(Env e, Value w, Value v)
		{
			return w.EqualTo(e, v);
		}
		public static Value IsMatchV(Env e, Value w, Value v)
		{
			return new Num(IsMatch(e, w, v) ? 1 : 0);
		}
		public static Value IsNotMatchV(Env e, Value w, Value v)
		{
			return new Num(IsMatch(e, w, v) ? 0 : 1);
		}
		public static Value Eq(Env e, Value w, Value v)
		{
			return new Num(EqualTo(e, w, v) ? 1 : 0);
		}
		public static Value Ne(Env e, Value w, Value v)
		{
			return new Num(EqualTo(e, w, v) ? 0 : 1);
		}
		public static Value Lt(Env e, Value w, Value v)
		{
			return new Num(CompareTo(e, w, v) < 0 ? 1 : 0);
		}
		public static Value Le(Env e, Value w, Value v)
		{
			return new Num(CompareTo(e, w, v) <= 0 ? 1 : 0);
		}
		public static Value Gt(Env e, Value w, Value v)
		{
			return new Num(CompareTo(e, w, v) > 0 ? 1 : 0);
		}
		public static Value Ge(Env e, Value w, Value v)
		{
			return new Num(CompareTo(e, w, v) >= 0 ? 1 : 0);
		}
		public static Value Concat(Env e, Value w, Value v)
		{
			return w.Concat(e, v);
		}
		public static Value Add(Env e, Value w, Value v)
		{
			return w.Add(e, v);
		}
		public static Value Sub(Env e, Value w, Value v)
		{
			return w.Sub(e, v);
		}
		public static Value Mul(Env e, Value w, Value v)
		{
			return w.Mul(e, v);
		}
		public static Value Div(Env e, Value w, Value v)
		{
			return w.Div(e, v);
		}
		public static Value Mod(Env e, Value w, Value v)
		{
			return w.Mod(e, v);
		}
		public static Value Neg(Env e, Value w)
		{
			return w.Neg(e);
		}
		public static Value Pos(Env e, Value w)
		{
			return w.Pos(e);
		}
		public static Value Pow(Env e, Value w, Value v)
		{
			return w.Pow(e, v);
		}
		public static Value Field(Env e, Value w)
		{
			return w.Field(e);
		}
		public static Value Not(Env e, Value w)
		{
			return w.Not(e);
		}

		public static string Varname(Value v1)
		{
			if(!(v1 is Sym))
			{
				throw new AwkRuntimeException("not lvalue");
			}
			return ((Sym)v1).Name;
		}
	}

	public class Num : Value
	{
		private double val;
		public Num(double v)
		{
			val = v;
		}
		public override bool ToBool(Env e)
		{
			return val == 0 ? false : true;
		}
		public override bool ToBoolMatch(Env e)
		{
			return val == 0 ? false : true;
		}
		public override string ToStr(Env e)
		{
			return AwkPrintf.Printf(e, e.GetConvfmt(), new Num(val));
		}
		public override double ToNum(Env e)
		{
			return val;
		}
		public override AwkRegex ToRE(Env e)
		{
			return new AwkRegex(ToStr(e));
		}
		public override Value ToRval(Env e)
		{
			return this;
		}
		public override bool IsEqualTo(Value v)
		{
			if(v is Num)
			{
				return val == ((Num)v).val;
			}
			return false;
		}
		public override bool IsNum(Env e)
		{
			return true;
		}
		public override string ToString()
		{
			return val.ToString();
		}
	}

	public class Str : Value
	{
		private string str;
		public Str(string s)
		{
			str = s;
		}
		public override bool ToBool(Env e)
		{
			Regex r = new Regex("^[\\+-]?([0-9]+(\\.[0-9]+)?|\\.[0-9]+)([eE][\\+-][0-9]+)?$");
			Match m = r.Match(str);
			if(m.Success)
			{
				return ToNum(e) == 0 ? false : true;
			}
			else
			{
				return str.Equals("") ? false : true;
			}
		}
		public override bool ToBoolMatch(Env e)
		{
			return str.Equals("") ? false : true;
		}
		public override string ToStr(Env e)
		{
			return str;
		}
		public override double ToNum(Env e)
		{
			Regex r = new Regex("^[\\+-]?([0-9]+(\\.[0-9]+)?|\\.[0-9]+)([eE][\\+-][0-9]+)?");
			Match m = r.Match(str);
			if(m.Success)
			{
				try
				{
					return double.Parse(m.Value);
				}
				catch(OverflowException)
				{
					if(str[0] == '-')
					{
						return double.NegativeInfinity;
					}
					else
					{
						return double.PositiveInfinity;
					}
				}
			}
			else
			{
				return 0;
			}
		}
		public override AwkRegex ToRE(Env e)
		{
			return new AwkRegex(str);
		}
		public override Value ToRval(Env e)
		{
			return this;
		}
		public override bool IsEqualTo(Value v)
		{
			if(v is Str)
			{
				return str.Equals(((Str)v).str);
			}
			return false;
		}
		public override bool IsNum(Env e)
		{
			Regex r = new Regex("^[\\+-]?([0-9]+(\\.[0-9]+)?|\\.[0-9]+)([eE][\\+-][0-9]+)?");
			Match m = r.Match(str);
			return m.Success;
		}
		public override string ToString()
		{
			return "\"" + str + "\"";
		}
	}

	public class ReE: Value
	{
		private AwkRegex pat;
		private string str;
		public ReE(string s)
		{
			pat = new AwkRegex(s);
			str = s;
		}
		public override bool ToBool(Env e)
		{
			return pat.IsMatch(e.GetField(0));
		}
		public override bool ToBoolMatch(Env e)
		{
			return pat.IsMatch(e.GetField(0));
		}
		public override string ToStr(Env e)
		{
			return pat.IsMatch(e.GetField(0)) ? "1" : "0";
		}
		public override double ToNum(Env e)
		{
			return pat.IsMatch(e.GetField(0)) ? 1 : 0;
		}
		public override AwkRegex ToRE(Env e)
		{
			return pat;
		}
		public override Value ToRval(Env e)
		{
			return new Num(ToNum(e));
		}
		public override bool IsEqualTo(Value v)
		{
			if(v is ReE)
			{
				return str.Equals(((ReE)v).str);
			}
			return false;
		}
		public override bool IsNum(Env e)
		{
			return false;
		}
		public override string ToString()
		{
			return "/" + str + "/";
		}
	}

	public class Sym : Value
	{
		public string Name { get; private set; }
		public Sym(string n)
		{
			Name = n;
		}
		public override bool ToBool(Env e)
		{
			return ToRval(e).ToBool(e);
		}
		public override bool ToBoolMatch(Env e)
		{
			return ToRval(e).ToBool(e);
		}
		public override string ToStr(Env e)
		{
			return ToRval(e).ToStr(e);
		}
		public override double ToNum(Env e)
		{
			return ToRval(e).ToNum(e);
		}
		public override AwkRegex ToRE(Env e)
		{
			return ToRval(e).ToRE(e);
		}
		public override Value ToRval(Env e)
		{
			if(Name.Equals("length"))
			{
				return new Num(e.GetField(0).Length);
			}
			else
			{
				return e.GetVariable(Name);
			}
		}
		public override bool IsEqualTo(Value v)
		{
			if(v is Sym)
			{
				return Name.Equals(((Sym)v).Name);
			}
			return false;
		}
		public override bool IsNum(Env e)
		{
			return ToRval(e).IsNum(e);
		}
		public override string ToString()
		{
			return Name;
		}
	}

	public interface AST
	{
		Value Evaluate(Env e);
	}

	public class Assign : AST
	{
		private AST left;
		private AST right;
		private Binary operate;
		public Assign(AST l, AST r, Binary op)
		{
			left = l;
			right = r;
			operate = op;
		}
		public static void SetVariable(AST lval, Env e, Value v)
		{
			List<Value> vs;
			ArrayVar arr;
			Value v1;
			if(lval is ConstVal)
			{
				v1 = lval.Evaluate(e);
				e.SetVariable(Value.Varname(v1), v);
			}
			else if(lval is ArrayVar)
			{
				arr = (ArrayVar)lval;
				vs = new List<Value>();
				foreach(AST a in arr.Args)
				{
					vs.Add(a.Evaluate(e));
				}
				e.SetArray(arr.Symbol.Name, vs, v);
			}
			else if(lval is OperateField)
			{
				v1 = ((OperateField)lval).Node.Evaluate(e);
				e.SetField((int)v1.ToNum(e), v);
			}
			else
			{
				throw new AwkRuntimeException("not lvalue");
			}
		}
		public Value Evaluate(Env e)
		{
			Value v1, v2;
			v2 = right.Evaluate(e).ToRval(e);
			if(operate != null)
			{
				v1 = left.Evaluate(e);
				v2 = operate(e, v1, v2);
			}
			Assign.SetVariable(left, e, v2);
			return v2;
		}
	}

	public class Cond : AST
	{
		private AST cond, then, nonsi;
		public Cond(AST c, AST i, AST e)
		{
			cond = c;
			then = i;
			nonsi = e;
		}
		public Value Evaluate(Env e)
		{
			if(cond.Evaluate(e).ToBool(e))
			{
				return then.Evaluate(e);
			}
			else
			{
				return nonsi.Evaluate(e);
			}
		}
	}

	public class LIor : AST
	{
		private AST left, right;
		public LIor(AST l, AST r)
		{
			left = l;
			right = r;
		}
		public Value Evaluate(Env e)
		{
			Value v1 = left.Evaluate(e);
			return v1.ToBool(e) ? new Num(1) : right.Evaluate(e);
		}
	}

	public class LAnd : AST
	{
		private AST left, right;
		public LAnd(AST l, AST r)
		{
			left = l;
			right = r;
		}
		public Value Evaluate(Env e)
		{
			Value v1 = left.Evaluate(e);
			return v1.ToBool(e) ? right.Evaluate(e) : new Num(0);
		}
	}

	public class InArray : AST
	{
		private string name;
		private List<AST> inval;
		public InArray(string n, AST a)
		{
			name = n;
			if(a is ExprGroup)
			{
				inval = ((ExprGroup)a).GetGroup();
			}
			else
			{
				inval = new List<AST>(new AST[] { a });
			}
		}
		public Value Evaluate(Env e)
		{
			List<Value> vs = new List<Value>();
			foreach(AST a in inval)
			{
				vs.Add(a.Evaluate(e));
			}
			return new Num(e.InArray(name, vs) ? 1 : 0);
		}
	}

	public class FileApp : AST
	{
		public AST Left { get; private set; }
		public AST Right { get; private set; }
		public FileApp(AST l, AST r)
		{
			Left = l;
			Right = r;
		}
		public Value Evaluate(Env e)
		{
			throw new AwkRuntimeException("Syntax error");
		}
	}

	public class OperateBinary : AST
	{
		public AST Left { get; private set; }
		public AST Right { get; private set; }
		public Binary Operate { get; private set; }
		public OperateBinary(AST l, AST r, Binary op)
		{
			Left = l;
			Right = r;
			Operate = op;
		}
		public Value Evaluate(Env e)
		{
			Value v1, v2;
			v1 = Left.Evaluate(e);
			v2 = Right.Evaluate(e);
			return Operate(e, v1, v2);
		}
	}

	public class OperateUnary : AST
	{
		private AST node;
		private Unary operate;
		public OperateUnary(AST x, Unary op)
		{
			node = x;
			operate = op;
		}
		public Value Evaluate(Env e)
		{
			Value v1;
			v1 = node.Evaluate(e);
			return operate(e, v1);
		}
	}

	public class IncDec : AST
	{
		private AST node;
		private Value num;
		private int prepost;
		public IncDec(AST n, int val, int pp)
		{
			node = n;
			num = new Num(val);
			prepost = pp;
		}
		public Value Evaluate(Env e)
		{
			double d1;
			Value v1, v2;
			v1 = node.Evaluate(e);
			d1 = v1.ToNum(e);
			v2 = Value.Add(e, v1, num);
			Assign.SetVariable(node, e, v2);
			return prepost > 0 ? v2 : new Num(d1);
		}
	}

	public class OperateField : AST
	{
		public AST Node { get; private set; }
		public OperateField(AST x)
		{
			Node = x;
		}
		public Value Evaluate(Env e)
		{
			Value v1;
			v1 = Node.Evaluate(e);
			return Value.Field(e, v1);
		}
	}

	public class ExprGroup : AST
	{
		private List<AST> gruppe;
		public ExprGroup(List<AST> a)
		{
			gruppe = new List<AST>(a);
		}
		public Value Evaluate(Env e)
		{
			if(gruppe.Count > 1)
			{
				throw new AwkRuntimeException("Syntax error");
			}
			return gruppe[0].Evaluate(e);
		}
		public List<AST> GetGroup()
		{
			return new List<AST>(gruppe);
		}
	}

	public class ConstVal : AST
	{
		private Value val;
		public ConstVal(Value v)
		{
			val = v;
		}
		public Value Evaluate(Env e)
		{
			return val;
		}
	}

	public class ArrayVar : AST
	{
		public Sym Symbol { get; private set; }
		public List<AST> Args { get; private set; }
		public ArrayVar(string n, List<AST> a)
		{
			Symbol = new Sym(n);
			Args = new List<AST>(a);
		}
		public Value Evaluate(Env e)
		{
			List<Value> vs = new List<Value>();
			foreach(AST a in Args)
			{
				vs.Add(a.Evaluate(e));
			}
			return e.GetArray(Symbol.Name, vs);
		}
	}

	public class CallFunction : AST
	{
		private List<AST> args;
		private string name;
		public CallFunction(string n, List<AST> a)
		{
			name = n;
			args = a;
		}
		public Value Evaluate(Env e)
		{
			List<Value> vs = new List<Value>();
			foreach(AST a in args)
			{
				vs.Add(a.Evaluate(e));
			}
			return e.CallFunction(name, vs);
		}
	}

	public class CallSplit : AST
	{
		private List<AST> args;
		public CallSplit(List<AST> a)
		{
			args = a;
		}
		public Value Evaluate(Env e)
		{
			string s1, nam;
			string[] res;
			AwkRegex r1;
			if(args.Count < 2 || args.Count > 3)
			{
				throw new AwkRuntimeException("split: invalid arity");
			}
			s1 = args[0].Evaluate(e).ToStr(e);
			r1 = args.Count > 2 ? args[2].Evaluate(e).ToRE(e) : new AwkRegex(e.GetFS());
			res = r1.Split(s1);
			nam = Value.Varname(args[1].Evaluate(e));
			e.ClearArray(nam);
			for(int i = 0; i < res.Length; i++)
			{
				e.SetArray(nam, new List<Value>(new Value[] { new Num(i + 1) }), new Str(res[i]));
			}
			return new Num(res.Length);
		}
	}

	public class CallSubGsub : AST
	{
		private List<AST> args;
		private bool glbl;
		public CallSubGsub(List<AST> a, bool g)
		{
			args = a;
			glbl = g;
		}
		private static string ParseReplacement(string rep, AwkMatch m)
		{
			StringBuilder res = new StringBuilder();
			bool esc = false;
			for(int i = 0; i < rep.Length; i++)
			{
				if(esc)
				{
					if(rep[i] != '&')
					{
						res.Append('\\');
					}
					res.Append(rep[i]);
					esc = false;
				}
				else
				{
					if(rep[i] == '\\')
					{
						esc = true;
					}
					else if(rep[i] == '&')
					{
						res.Append(m.Value);
					}
					else
					{
						res.Append(rep[i]);
					}
				}
			}
			return res.ToString();
		}
		private Value Asubgsub(Env e, AwkRegex r1, string s1, string sr, ref int len)
		{
			string rp, rs, rb;
			AwkMatch m;
			m = r1.Match(sr);
			if(!m.Success)
			{
				len = 0;
				return new Str(sr);
			}
			else
			{
				rp = ParseReplacement(s1, m);
				if(glbl)
				{
					rs = r1.Replace(sr, rp, -1, ref len);
				}
				else
				{
					rs = r1.Replace(sr, rp, 1, ref len);
					len = 1;
				}
				return new Str(rs);
			}
		}
		public Value Evaluate(Env e)
		{
			AwkRegex r1;
			string s1;
			Value rs;
			AST rp;
			int len = 0;
			if(args.Count < 2 || args.Count > 3)
			{
				throw new AwkRuntimeException((glbl ? "gsub" : "sub") + ": invalid arity");
			}
			r1 = args[0].Evaluate(e).ToRE(e);
			s1 = args[1].Evaluate(e).ToStr(e);
			if(args.Count > 2)
			{
				rp = args[2];
			}
			else
			{
				rp = new OperateField(new ConstVal(new Num(0)));
			}
			rs = Asubgsub(e, r1, s1, rp.Evaluate(e).ToStr(e), ref len);
			Assign.SetVariable(rp, e, rs);
			return new Num(len);
		}
	}

	public class Getline : AST
	{
		private AST expr;
		private AST redirect = null;
		public Getline(AST e, AST r)
		{
			expr = e;
			redirect = r;
		}
		public Value Evaluate(Env e)
		{
			try
			{
				AwkStream s;
				string l;
				s = redirect == null ? e.Stream : e.OpenReader(redirect.Evaluate(e).ToStr(e));
				if(s == null)
				{
					return new Num(-1);
				}
				l = s.ReadRecord(e);
				if(l == null)
				{
					return new Num(0);
				}
				if(expr == null)
				{
					e.SetField(0, new Str(l));
				}
				else
				{
					Assign.SetVariable(expr, e, new Str(l));
				}
				return new Num(1);
			}
			catch(IOException)
			{
				return new Num(-1);
			}
		}
	}

	public class VM
	{
		public const int NoJmp = -1;
		public const int Next = -100;
		public const int Exit = -101;
		private List<VMCode> codes = new List<VMCode>();
		private List<List<int>> breaks = new List<List<int>>();
		private List<List<int>> continues = new List<List<int>>();
		public int AddCode(VMCode code)
		{
			codes.Add(code);
			return codes.Count - 1;
		}
		public int AddCode(VMInst inst, Operand op)
		{
			return AddCode(new VMCode(inst, op));
		}
		public Operand GetOp(int addr)
		{
			return codes[addr].Op;
		}
		public int GetAddr()
		{
			return codes.Count;
		}
		public void BeginLoop()
		{
			breaks.Add(new List<int>());
			continues.Add(new List<int>());
		}
		public void AddBreak(int addr)
		{
			breaks[breaks.Count - 1].Add(addr);
		}
		public void AddContinue(int addr)
		{
			continues[continues.Count - 1].Add(addr);
		}
		public void EndLoop(int baddr, int caddr)
		{
			foreach(int a in breaks[breaks.Count - 1])
			{
				codes[a].Op.Addr = baddr;
			}
			foreach(int a in continues[continues.Count - 1])
			{
				codes[a].Op.Addr = caddr;
			}
			breaks.RemoveAt(breaks.Count - 1);
			continues.RemoveAt(continues.Count - 1);
		}
		public int Execute(Env e, AwkRuntime run)
		{
			int pc = 0, ret;
			while(pc < codes.Count)
			{
				ret = codes[pc].Execute(e, run);
				if(ret >= 0)
				{
					pc = ret;
				}
				else if(ret == VM.NoJmp)
				{
					pc++;
				}
				else if(ret == VM.Next)
				{
					return VM.Next;
				}
			}
			return 0;
		}
	}

	public class VMCode
	{
		private VMInst inst;
		public Operand Op { get; private set; }
		public VMCode(VMInst i, Operand op)
		{
			inst = i;
			Op = op;
		}
		public int Execute(Env e, AwkRuntime run)
		{
			return inst(e, run, Op);
		}
	}

	public class Operand
	{
		public int Addr { get; set; }
		public AST Expr { get; private set; }
		public string Name { get; private set; }
		public SPrint Print { get; private set; }
		public Operand(AST e, int a, string n)
		{
			Addr = a;
			Expr = e;
			Name = n;
		}
		public Operand(AST e, int a) : this(e, a, null)
		{
		}
		public Operand(int a) : this(null, a, null)
		{
		}
		public Operand(AST e) : this(e, -1, null)
		{
		}
		public Operand(string n) : this(null, -1, n)
		{
		}
		public Operand(SPrint p)
		{
			Print = p;
		}
	}

	public class AwkRuntime
	{
		private List<List<Value>> eachstack = new List<List<Value>>();
		private HashSet<Pattern> matching = new HashSet<Pattern>();
		public bool Begin { get; set; }
		public bool End { get; set; }
		public void PushEach(List<Value> vals)
		{
			eachstack.Add(new List<Value>(vals));
		}
		public Value PopEachValue()
		{
			List<Value> vals = eachstack[eachstack.Count - 1];
			Value val;
			if(vals.Count > 0)
			{
				val = vals[0];
				vals.RemoveAt(0);
				return val;
			}
			else
			{
				eachstack.RemoveAt(eachstack.Count - 1);
				return null;
			}
		}
		public bool IsMatching(Pattern pat)
		{
			return matching.Contains(pat);
		}
		public void SetMatching(Pattern pat, bool stat)
		{
			if(stat && !matching.Contains(pat))
			{
				matching.Add(pat);
			}
			else if(!stat && matching.Contains(pat))
			{
				matching.Remove(pat);
			}
		}
	}

	public delegate int VMInst(Env e, AwkRuntime run, Operand op);
	public class VMInstSet
	{
		public static int Expr(Env e, AwkRuntime run, Operand op)
		{
			op.Expr.Evaluate(e);
			return VM.NoJmp;
		}
		public static int Jmp(Env e, AwkRuntime run, Operand op)
		{
			return op.Addr;
		}
		public static int Jnz(Env e, AwkRuntime run, Operand op)
		{
			return op.Expr.Evaluate(e).ToBool(e) ? op.Addr : VM.NoJmp;
		}
		public static int Jz(Env e, AwkRuntime run, Operand op)
		{
			return op.Expr.Evaluate(e).ToBool(e) ? VM.NoJmp : op.Addr;
		}
		public static int InitEach(Env e, AwkRuntime run, Operand op)
		{
			run.PushEach(e.GetArrays(op.Name));
			return VM.NoJmp;
		}
		public static int Each(Env e, AwkRuntime run, Operand op)
		{
			AST a;
			Value v;
			v = run.PopEachValue();
			if(v == null)
			{
				return op.Addr;
			}
			else
			{
				a = new Assign(op.Expr, new ConstVal(v), null);
				a.Evaluate(e);
				return VM.NoJmp;
			}
		}
		public static int Next(Env e, AwkRuntime run, Operand op)
		{
			return VM.Next;
		}
		public static int Exit(Env e, AwkRuntime run, Operand op)
		{
			if(op.Expr == null)
			{
				Environment.Exit(0);
			}
			else
			{
				Environment.Exit((int)op.Expr.Evaluate(e).ToNum(e));
			}
			return VM.NoJmp;
		}
		public static int Print(Env e, AwkRuntime run, Operand op)
		{
			op.Print.Print(e);
			return VM.NoJmp;
		}
		public static int Nop(Env e, AwkRuntime run, Operand op)
		{
			return VM.NoJmp;
		}
	}

	public interface ASTS
	{
		void CompileVM(VM v);
	}

	public class SIf : ASTS
	{
		private AST scond;
		private ASTS sif, selse;
		public SIf(AST c, ASTS i, ASTS e)
		{
			scond = c;
			sif = i;
			selse = e;
		}
		public SIf(AST c, ASTS i) : this(c, i, null)
		{
		}
		public void CompileVM(VM v)
		{
			int l1, l2;
			l1 = v.AddCode(VMInstSet.Jz, new Operand(scond, -2));
			sif.CompileVM(v);
			if(selse != null)
			{
				l2 = v.AddCode(VMInstSet.Jmp, new Operand(-2));
				v.GetOp(l1).Addr = v.GetAddr();
				selse.CompileVM(v);
				v.GetOp(l2).Addr = v.GetAddr();
			}
			else
			{
				v.GetOp(l1).Addr = v.GetAddr();
			}
		}
	}

	public class SWhile : ASTS
	{
		private AST scond;
		private ASTS sbody;
		public SWhile(AST c, ASTS b)
		{
			scond = c;
			sbody = b;
		}
		public void CompileVM(VM v)
		{
			int l1;
			v.BeginLoop();
			l1 = v.AddCode(VMInstSet.Jz, new Operand(scond, -2));
			sbody.CompileVM(v);
			v.AddCode(VMInstSet.Jmp, new Operand(l1));
			v.GetOp(l1).Addr = v.GetAddr();
			v.EndLoop(v.GetAddr(), l1);
		}
	}

	public class SDo : ASTS
	{
		private AST scond;
		private ASTS sbody;
		public SDo(AST c, ASTS b)
		{
			scond = c;
			sbody = b;
		}
		public void CompileVM(VM v)
		{
			int l1;
			v.BeginLoop();
			l1 = v.GetAddr();
			sbody.CompileVM(v);
			v.AddCode(VMInstSet.Jnz, new Operand(scond, l1));
			v.EndLoop(v.GetAddr(), l1);
		}
	}

	public class SFor : ASTS
	{
		private AST sinit, scond, snext;
		private ASTS sbody;
		public SFor(AST i, AST c, AST n, ASTS b)
		{
			sinit = i;
			scond = c;
			snext = n;
			sbody = b;
		}
		public void CompileVM(VM v)
		{
			int l1, l2;
			v.BeginLoop();
			if(sinit != null)
			{
				v.AddCode(VMInstSet.Expr, new Operand(sinit));
			}
			if(scond != null)
			{
				l1 = v.AddCode(VMInstSet.Jz, new Operand(scond, -2));
			}
			else
			{
				l1 = v.AddCode(VMInstSet.Nop, new Operand(-2));
			}
			sbody.CompileVM(v);
			if(snext != null)
			{
				l2 = v.AddCode(VMInstSet.Expr, new Operand(snext));
			}
			else
			{
				l2 = v.AddCode(VMInstSet.Nop, new Operand(-2));
			}
			v.AddCode(VMInstSet.Jmp, new Operand(l1));
			if(scond != null)
			{
				v.GetOp(l1).Addr = v.GetAddr();
			}
			v.EndLoop(v.GetAddr(), l2);
		}
	}

	public class SForeach : ASTS
	{
		private AST slval;
		private string sname;
		private ASTS sbody;
		public SForeach(AST l, string n, ASTS b)
		{
			slval = l;
			sname = n;
			sbody = b;
		}
		public void CompileVM(VM v)
		{
			int l1, l2;
			v.BeginLoop();
			v.AddCode(VMInstSet.InitEach, new Operand(sname));
			l1 = v.AddCode(VMInstSet.Each, new Operand(slval, -2));
			sbody.CompileVM(v);
			l2 = v.AddCode(VMInstSet.Jmp, new Operand(l1));
			v.GetOp(l1).Addr = v.GetAddr();
			v.EndLoop(v.GetAddr(), l2);
		}
	}

	public class SBreak : ASTS
	{
		public void CompileVM(VM v)
		{
			int l1;
			l1 = v.AddCode(VMInstSet.Jmp, new Operand(-2));
			v.AddBreak(l1);
		}
	}

	public class SContinue : ASTS
	{
		public void CompileVM(VM v)
		{
			int l1;
			l1 = v.AddCode(VMInstSet.Jmp, new Operand(-2));
			v.AddContinue(l1);
		}
	}

	public class SNext : ASTS
	{
		public void CompileVM(VM v)
		{
			v.AddCode(VMInstSet.Next, new Operand(-2));
		}
	}

	public class SExit : ASTS
	{
		private AST expr;
		public SExit(AST e)
		{
			expr = e;
		}
		public SExit() : this(null)
		{
		}
		public void CompileVM(VM v)
		{
			v.AddCode(VMInstSet.Exit, new Operand(expr));
		}
	}

	public class SBlock : ASTS
	{
		private List<ASTS> block;
		public SBlock(List<ASTS> b)
		{
			block = new List<ASTS>(b);
		}
		public void CompileVM(VM v)
		{
			foreach(ASTS s in block)
			{
				s.CompileVM(v);
			}
		}
	}

	public class SExpr : ASTS
	{
		private AST expr;
		public SExpr(AST e)
		{
			expr = e;
		}
		public void CompileVM(VM v)
		{
			v.AddCode(VMInstSet.Expr, new Operand(expr));
		}
	}

	public class SPrint : ASTS
	{
		private List<AST> exprs;
		private AST redirect = null;
		private bool append;
		private bool printf;
		public SPrint(List<AST> e, bool f)
		{
			printf = f;
			exprs = new List<AST>();
			for(int i = 0; i < e.Count; i++)
			{
				if(i < e.Count - 1)
				{
					exprs.Add(e[i]);
				}
				else if(e[i] is FileApp)
				{
					exprs.Add(((FileApp)e[i]).Left);
					redirect = ((FileApp)e[i]).Right;
					append = true;
				}
				else if(e[i] is OperateBinary && ((OperateBinary)e[i]).Operate == Value.Gt)
				{
					exprs.Add(((OperateBinary)e[i]).Left);
					redirect = ((OperateBinary)e[i]).Right;
					append = false;
				}
				else
				{
					exprs.Add(e[i]);
				}
			}
		}
		public void CompileVM(VM v)
		{
			v.AddCode(VMInstSet.Print, new Operand(this));
		}
		public void Print(Env e)
		{
			StringBuilder put = new StringBuilder();
			StreamWriter wrt;
			string ofs = e.GetOFS();
			string ors = e.GetORS();
			string dlm = "";
			string fmt;
			Value[] vs;
			Value v;
			if(printf)
			{
				vs = new Value[exprs.Count - 1];
				fmt = exprs[0].Evaluate(e).ToStr(e);
				for(int i = 1; i < exprs.Count; i++)
				{
					vs[i - 1] = exprs[i].Evaluate(e);
				}
				put.Append(AwkPrintf.Printf(e, fmt, vs));
			}
			else
			{
				if(exprs.Count > 0)
				{
					foreach(AST a in exprs)
					{
						put.Append(dlm);
						v = a.Evaluate(e);
						if(v is Num)
						{
							v = new Str(AwkPrintf.Printf(e, e.GetOfmt(), v));
						}
						put.Append(v.ToStr(e));
						dlm = ofs;
					}
				}
				else
				{
					put.Append(e.GetField(0));
				}
				put.Append(ors);
			}
			if(redirect == null)
			{
				Console.Write(put.ToString().Replace("\n", Environment.NewLine));
			}
			else
			{
				wrt = e.OpenWriter(redirect.Evaluate(e).ToStr(e), append);
				wrt.Write(put.ToString().Replace("\n", Environment.NewLine));
			}
		}
	}

	public interface Pattern
	{
		bool Match(Env e, AwkRuntime run);
	}

	public class PBegin : Pattern
	{
		public bool Match(Env e, AwkRuntime run)
		{
			return run.Begin;
		}
	}

	public class PEnd : Pattern
	{
		public bool Match(Env e, AwkRuntime run)
		{
			return run.End;
		}
	}

	public class PEmpty : Pattern
	{
		public bool Match(Env e, AwkRuntime run)
		{
			return !run.Begin && !run.End;
		}
	}

	public class PExpr : Pattern
	{
		private AST expr;
		public PExpr(AST e)
		{
			expr = e;
		}
		public bool Match(Env e, AwkRuntime run)
		{
			return !run.Begin && !run.End && expr.Evaluate(e).ToBoolMatch(e);
		}
	}

	public class PRange : Pattern
	{
		private AST efrom, eto;
		public PRange(AST f, AST t)
		{
			efrom = f;
			eto = t;
		}
		public bool Match(Env e, AwkRuntime run)
		{
			bool mt;
			if(run.Begin || run.End)
			{
				return false;
			}
			if(run.IsMatching(this))
			{
				run.SetMatching(this, !eto.Evaluate(e).ToBoolMatch(e));
				return true;
			}
			else
			{
				mt = efrom.Evaluate(e).ToBool(e);
				run.SetMatching(this, mt);
				if(mt)
				{
					run.SetMatching(this, !eto.Evaluate(e).ToBoolMatch(e));
				}
				return mt;
			}
		}
	}

	public class AwkAction
	{
		private Pattern pat;
		private VM code;
		public AwkAction(Pattern p, ASTS s)
		{
			pat = p;
			code = new VM();
			s.CompileVM(code);
		}
		public AwkAction(Pattern p)
		{
			pat = p;
			code = null;
		}
		public bool IsBegin()
		{
			return pat is PBegin;
		}
		public bool Execute(Env e, AwkRuntime run)
		{
			if(pat.Match(e, run))
			{
				if(code != null)
				{
					if(code.Execute(e, run) == VM.Next)
					{
						return true;
					}
				}
				else
				{
					Console.WriteLine(e.GetField(0));
				}
			}
			return false;
		}
	}

	public class AwkStream
	{
		private TextReader reader;
		private string line = "";
		private int ptr = 0;
		private bool closed = false;
		public AwkStream(TextReader r)
		{
			reader = r;
		}
		private string ReadLine()
		{
			if(closed || line == null)
			{
				return null;
			}
			else if(reader == null)
			{
				return Console.ReadLine();
			}
			else
			{
				return reader.ReadLine();
			}
		}
		private int Getchar()
		{
			if(line == null)
			{
				return -1;
			}
			else if(ptr < line.Length)
			{
				return line[ptr++];
			}
			else
			{
				line = ReadLine();
				ptr = 0;
				return line != null ? '\n' : -1;
			}
		}
		public string ReadRecord(Env e)
		{
			StringBuilder b;
			string l;
			int c, d;
			d = e.GetRS();
			if(d == '\n')
			{
				line = ReadLine();
				return line;
			}
			else if(d < 0)
			{
				b = new StringBuilder();
				if(line == null)
				{
					return null;
				}
				while((line = ReadLine()) != null && line.Equals(""))
				{
					/* do nothing */
				}
				if(line == null)
				{
					return null;
				}
				b.Append(line);
				while((line = ReadLine()) != null && !line.Equals(""))
				{
					b.Append("\n").Append(line);
				}
				return b.ToString();
			}
			else
			{
				b = new StringBuilder();
				while((c = Getchar()) >= 0 && c != d)
				{
					b.Append((char)c);
				}
				l = b.ToString();
				return l.Equals("") ? null : l;
			}
		}
		public void Close()
		{
			if(reader != null && !closed)
			{
				reader.Close();
				closed = true;
			}
		}
	}
}